import logging
import traceback

import six

from ..conf import settings
from ..storage import PageHistory
from ..bash import execute_bash
from .utils import bake_parametrized


PYTHON_ERROR = "transform.python must set global variables ok and content"
logger = logging.getLogger(__name__)


def changes_transform_factory(value, conf):
    style = value.lower() if value else None
    return PageHistory(conf, style=style).report_changes


def python_transform(code, content, conf):
    logger.info("Python transform")
    logger.debug(code)
    assert 'content' in code, PYTHON_ERROR
    try:
        namespace = {'ok': True, 'content': content}
        exec(code, {'creds': settings().creds, 'conf': conf}, namespace)
        return namespace['ok'], six.text_type(namespace['content'])
    except:
        logger.exception("Python transform raised an Exception")
        return False, traceback.format_exc()


def bash_transform(code, content):
    return execute_bash(code, content)


def register():
    return {
        'changes': changes_transform_factory,
        'python': bake_parametrized(python_transform, pass_conf=True),
        'bash': bake_parametrized(bash_transform),
    }



import logging

from ..conf import settings


logger = logging.getLogger(__name__)
NAME = 'python'


class PythonNotify(object):

    def __init__(self, conf, value):
        self.code = value
        self.context = {
            'conf': conf,
            'creds': settings().creds,
        }

    def __call__(self, report):
        context = dict(
            self.context,
            text=report,  # legacy
            content=report,
        )
        logger.info("Executing Python notifier")
        logger.debug(self.code)
        exec(self.code, context)


def notify_factory(conf, value):
    return PythonNotify(conf, value)



from .webhook import WebHookNotify, webhook_factory

from ..conf import settings


class MailgunNotify(WebHookNotify):

    CREDS_KEY = 'mailgun'

    def __init__(self, conf, value, **kwargs):
        self.mailgun_creds = settings().creds.get('mailgun', {})
        self.mailgun_creds.update(value or {})
        domain = self.mailgun_creds['domain']
        self.context = {
            'subject': 'Kibitzr update for ' + conf['name'],
            'from': 'Kibitzr <mailgun@{0}>'.format(domain),
            'to': [self.mailgun_creds['to']],
        }
        super(MailgunNotify, self).__init__(conf=conf, value=value, **kwargs)

    def load_url(self, creds_key, value):
        return ('https://api.mailgun.net/v3/{0}/messages'
                .format(self.mailgun_creds['domain']))

    def configure_session(self):
        self.session.auth = ('api', self.mailgun_creds['key'])

    def payload(self, report):
        return dict(
            self.context,
            text=report,
        )


notify_factory = webhook_factory(MailgunNotify)



import tempfile

from kibitzr.notifier.bash import notify_factory as bash_factory


def test_bash_unicode_is_handled():
    content = u"\U0001F4A9"
    with tempfile.NamedTemporaryFile() as outfile:
        bash_notify = bash_factory({}, 'tee %s' % outfile.name)
        bash_notify(content)
        written = outfile.read().decode("utf-8")
        assert written == content


def test_bash_error_is_captured():
    bash_notify = bash_factory({}, 'command-not-found')
    bash_notify("hello")



import logging
import collections
from time import sleep

import requests
from cachecontrol import CacheControl
from kibitzr import __version__ as version


logger = logging.getLogger(__name__)


class SessionFetcher(object):
    RETRIABLE_EXCEPTIONS = (
        (requests.HTTPError, 5),
        (requests.ConnectionError, 15),
        (requests.Timeout, lambda retry: 60 * (retry + 1)),
    )
    # Explicitly listing exceptions from above to make pylint happy:
    EXCEPTED = (
        requests.HTTPError,
        requests.ConnectionError,
        requests.Timeout,
    )

    def __init__(self, conf):
        self.conf = conf
        self.session = CacheControl(requests.Session())
        self.session.headers.update({
            'User-agent': 'Kibitzr/' + version,
        })
        self.url = conf['url']
        self.valid_http = set(conf.get('valid_http', [200]))

    def fetch(self):
        retries = 3
        for retry in range(retries):
            try:
                response = self.session.get(self.url)
            except self.EXCEPTED as exc:
                if retry < retries - 1:
                    self.sleep_on_exception(exc, retry)
                else:
                    raise
            else:
                ok = (response.status_code in self.valid_http)
                text = response.text
                return ok, text

    def sleep_on_exception(self, exc, retry):
        for klass, seconds in self.RETRIABLE_EXCEPTIONS:
            if isinstance(exc, klass):
                if isinstance(seconds, collections.Callable):
                    seconds = seconds(retry)
                sleep(seconds)
                break


def requests_fetcher(conf):
    def fetcher(conf):
        return session_fetcher.fetch()
    session_fetcher = SessionFetcher(conf)
    return fetcher



from .factory import fetcher_factory  # noqa


def persistent_firefox():
    from .browser.fetcher import persistent_firefox as real_persistent_firefox
    real_persistent_firefox()


def cleanup_fetchers():
    from .browser.launcher import cleanup
    cleanup()



"""
Built-in transforms
"""
import os
import pkgutil
import importlib
import logging

import six


logger = logging.getLogger(__name__)


def transform_factory(conf):
    return TransformPipeline(conf)


def load_transforms():
    path = os.path.dirname(os.path.abspath(__file__))
    before, sep, _ = __name__.rpartition('.')
    prefix = before + sep
    registry = {}
    for _, modname, _ in pkgutil.walk_packages([path], prefix):
        submodule = importlib.import_module(modname, __name__)
        if hasattr(submodule, 'register'):
            registry.update(submodule.register())
    return registry


class TransformPipeline(object):

    REGISTRY = load_transforms()

    def __init__(self, conf):
        self.conf = conf
        rules = self.conf.get('transform', [])
        if isinstance(rules, six.string_types):
            rules = [rules]
        self.transforms = [
            self.create_transform(rule)
            for rule in rules
        ]

    def run_pipeline(self, ok, content):
        for transform in self.transforms:
            if ok:
                ok, content = transform(content)
            else:
                break
        if not ok:
            content = self.on_error(content)
        if content:
            content = content.strip()
        return ok, content
    __call__ = run_pipeline

    def create_transform(self, rule):
        try:
            name, value = next(iter(rule.items()))
        except AttributeError:
            name, value = rule, None
        try:
            return self.REGISTRY[name](value, conf=self.conf)
        except KeyError:
            raise RuntimeError(
                "Unknown transform: %r" % (name,)
            )

    def on_error(self, content):
        error_policy = self.conf.get('error', 'notify')
        if error_policy == 'ignore':
            if content:
                logger.error("Ignoring error in %s",
                             repr(content)[:60])
            return None
        elif error_policy == 'notify':
            logger.debug("Notifying on error")
            return content
        else:
            logger.warning("Unknown error policy: %r", error_policy)
            logger.info("Defaulting to 'notify'")
            return content



import os
import logging
import tempfile
import contextlib


logger = logging.getLogger(__name__)


def execute_bash(code, stdin=None):
    if os.name == 'nt':
        executor = WindowsExecutor
    else:
        executor = BashExecutor
    return executor(code).execute(stdin)


class BashExecutor(object):

    def __init__(self, code):
        self.code = code

    def execute(self, stdin=None):
        stdin = self.prepare_input(stdin)
        with self.temp_file() as filename:
            ok, result = self.run_scipt(filename, stdin)
        return self.make_report(ok, result)

    @staticmethod
    def prepare_input(stdin):
        if stdin is not None:
            if not stdin.strip():
                logger.info("Skipping execution with empty content")
                return True, stdin
            stdin = stdin.encode("utf-8")
        return stdin

    @contextlib.contextmanager
    def temp_file(self):
        with tempfile.NamedTemporaryFile() as fp:
            logger.debug("Saving code to %r", fp.name)
            fp.write(self.code.encode('utf-8'))
            fp.flush()
            yield fp.name

    @staticmethod
    def run_scipt(name, stdin):
        from kibitzr.compat import sh
        logger.debug("Launching script %r", name)
        try:
            return True, sh.Command("bash")(name, _in=stdin)
        except sh.ErrorReturnCode as exc:
            return False, exc

    @staticmethod
    def make_report(ok, result):
        stdout = result.stdout.decode('utf-8')
        stderr = result.stderr.decode('utf-8')
        if ok:
            log = logger.debug
            report = stdout
        else:
            log = logger.error
            report = stderr
        log("Bash exit_code: %r", result.exit_code)
        log("Bash stdout: %s", stdout)
        log("Bash stderr: %s", stderr)
        return ok, report


class WindowsExecutor(BashExecutor):

    @contextlib.contextmanager
    def temp_file(self):
        with tempfile.NamedTemporaryFile(suffix='.bat', delete=False) as fp:
            try:
                logger.debug("Saving code to %r", fp.name)
                fp.write(self.code.encode('utf-8'))
                fp.close()
                yield fp.name
            finally:
                os.remove(fp.name)

    @staticmethod
    def run_scipt(name, stdin):
        from kibitzr.compat import sh
        logger.debug("Launching script %r", name)
        try:
            return True, sh.Command("cmd.exe")("/Q", "/C", name, _in=stdin)
        except sh.ErrorReturnCode as exc:
            return False, exc

    @staticmethod
    def make_report(ok, result):
        stdout = result.stdout
        stderr = result.stderr
        if ok:
            log = logger.debug
            report = stdout
        else:
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')
            log = logger.error
            report = stderr
        log("CMD stdout: %s", stdout)
        log("CMD stderr: %s", stderr)
        return ok, report



import os
import time
import shutil
import logging
import collections

from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)
from jinja2 import Template

from kibitzr.conf import settings
from .launcher import firefox, PROFILE_DIR


logger = logging.getLogger(__name__)
HOME_PAGE = 'https://kibitzr.github.io/'


def persistent_firefox():
    if not os.path.exists(PROFILE_DIR):
        os.makedirs(PROFILE_DIR)
    with firefox(headless=False) as driver:
        driver.get(HOME_PAGE)
        while True:
            try:
                # Property raises when browser is closed:
                driver.title
            except:
                # All kinds of things happen when closing Firefox
                break
            else:
                time.sleep(0.2)
        update_profile(driver)


def update_profile(driver):
    if os.path.exists(PROFILE_DIR):
        shutil.rmtree(PROFILE_DIR)
    shutil.copytree(
        driver.profile.profile_dir,
        PROFILE_DIR,
        ignore=shutil.ignore_patterns(
            "parent.lock",
            "lock",
            ".parentlock",
            "*.sqlite-shm",
            "*.sqlite-wal",
        ),
    )


def firefox_fetcher(conf):
    headless = conf.get('headless', True)
    with firefox(headless) as driver:
        fetcher = FirefoxFetcher(driver)
        return fetcher.fetch(conf)


class FirefoxFetcher(object):

    def __init__(self, driver):
        self.driver = driver

    def fetch(self, conf):
        """
        1. Fetch URL
        2. Run automation.
        3. Return HTML.
        4. Close the tab.
        """
        url = conf['url']
        self.driver.get(url)
        self._run_automation(conf)
        html = self._get_html()
        self._close_tab()
        return True, html

    def _run_automation(self, conf):
        """
        1. Fill form.
        2. Run scenario.
        3. Delay.
        """
        self._fill_form(self._find_form(conf))
        self._run_scenario(conf)
        self._delay(conf)

    @staticmethod
    def _fill_form(form):
        """
        Fill all inputs with provided Jinja2 templates.
        If no field had click key, submit last element.
        """
        clicked = False
        last_element = None
        for field in form:
            if field['text']:
                field['element'].clear()
                field['element'].send_keys(field['text'])
            if field['click']:
                field['element'].click()
                clicked = True
            last_element = field['element']
        if last_element:
            if not clicked:
                last_element.submit()

    def _find_form(self, conf):
        """
        Find elements defined in conf['form'].
        Render all Jinja2 templates from field['value'].
        Save all field['click'] triggers.
        If all found, return them as a list of dictionaries.
        If at least one was not found, return empty list.
        """
        form = conf.get('form', [])
        fields = []
        creds = settings().creds
        for field in form:
            click = field.get('click')
            text = self._parse_field_text(field, conf, creds)
            selector_type, selector = self._parse_field_selector(field)
            if selector:
                element = self._find_element(selector, selector_type)
                if element:
                    fields.append({
                        'element': element,
                        'text': text,
                        'click': click,
                    })
                else:
                    logging.warning("Element {%s: %s} not found",
                                    selector_type, selector)
        if len(fields) == len(form):
            return fields
        else:
            logging.info(
                "Skipped form filling because not all fields were found"
            )
            return []

    @staticmethod
    def _parse_field_selector(field):
        """
        Check field keys: css, xpath and id.
        Return the first found along with it's value.
        """
        for selector_type in ('css', 'xpath', 'id'):
            selector = field.get(selector_type)
            if selector:
                return selector_type, selector
        logger.warning("Form field does not define any selector "
                       "(id, css and xpath available): %r",
                       field)
        return None, None

    @staticmethod
    def _parse_field_text(field, conf, creds):
        """
        Return form field value from 3 options:
        1. If value key is present, render it as a Jinja2 template.
           The template has access to ``conf`` and ``creds`` dictionaries.
        2. If creds key is present, use it as a dot-delimited path
           in creds dictionary (e.g. ``account.login``).
        3. Otherwise return None
        """
        value = field.get('value')
        if value:
            template = Template(value)
            return template.render(
                conf=conf,
                creds=creds,
            )
        creds_path = field.get('creds')
        if creds_path:
            bread_crumbs = creds_path.split('.')
            node = creds
            for crumb in bread_crumbs:
                node = node[crumb]
            return node

    def _run_scenario(self, conf):
        scenario = conf.get('scenario')
        if isinstance(scenario, collections.Mapping):
            code = scenario['python']
            elements = scenario.get('elements', {})
        else:
            code = scenario
            elements = {}
        if elements:
            element_objects = self._find_elements(elements)
        else:
            element_objects = {}
        if code:
            self._exec_scenario(code, conf, element_objects)

    @staticmethod
    def _delay(conf):
        delay = conf.get('delay')
        if delay:
            time.sleep(delay)

    def _get_html(self):
        for attempt in range(3):
            try:
                elem = self.driver.find_element_by_xpath("//*")
                html = elem.get_attribute("outerHTML")
            except (NoSuchElementException, StaleElementReferenceException):
                # Crazy (but stable) race condition,
                # new page loaded after call to find_element_by_xpath
                # Just retry:
                if attempt < 2:
                    time.sleep(1)
                else:
                    raise
            else:
                return html

    def _find_elements(self, elements):
        logger.info("Finding elements")
        result = {}
        for key, selector in elements.items():
            name, _, selector_type = key.partition("|")
            result[name] = self._find_element(selector, selector_type.lower())
        return result

    def _find_element(self, selector, selector_type):
        """
        Return first matching displayed element of non-zero size
        or None if nothing found
        """
        if selector_type == 'css':
            elements = self.driver.find_elements_by_css_selector(selector)
        elif selector_type == 'xpath':
            elements = self.driver.find_elements_by_xpath_selector(selector)
        elif selector_type == 'id':
            elements = self.driver.find_elements_by_css_selector('#' + selector)
        else:
            raise RuntimeError(
                "Unknown selector_type: %s for selector: %s"
                % (selector_type, selector)
            )
        for element in elements:
            if element.is_displayed():
                if sum(element.size.values()) > 0:
                    return element

    def _exec_scenario(self, code, conf, elements):
        logger.info("Executing custom scenario")
        logger.debug(code)
        exec(
            code,
            {
                'conf': conf,
                'creds': settings().creds,
                'driver': self.driver,
                'elements': elements,
            },
        )

    def _close_tab(self):
        """
        Create a new tab and close the old one
        to avoid idle page resource usage
        """
        old_tab = self.driver.current_window_handle
        self.driver.execute_script('''window.open("about:blank", "_blank");''')
        self.driver.switch_to.window(old_tab)
        self.driver.close()
        self.driver.switch_to.window(self.driver.window_handles[0])



import pytest

from kibitzr.notifier.factory import CompositeNotifier

from ...compat import mock
from ...helpers import SettingsMock


@pytest.fixture()
def settings():
    """Override native settings singleton with empty one"""
    return SettingsMock.instance()


def test_python_unicode_is_handled(settings):
    content = u"\U0001F4A9"
    conf = {
        'notify': [{
            'python': 'conf["content"] = content * 2',
        }],
    }
    notify_func = CompositeNotifier(conf)
    notify_func(content)
    written = conf['content']
    assert written == content * 2


@mock.patch("kibitzr.notifier.factory.logger")
def test_exception_is_captured(fake_logger):
    notify_func = CompositeNotifier({'notify': [{'python': 'print(1 / 0)'}]})
    notify_func("bang")
    fake_logger.exception.assert_called_once_with(
        "Exception occurred during sending notification"
    )



import os
import logging
import pkgutil
import importlib


logger = logging.getLogger(__name__)


def dummy_notify_factory(notify_func):
    def factory(conf, value):
        return notify_func
    return factory


def load_notifiers():
    path = os.path.dirname(os.path.abspath(__file__))
    before, sep, _ = __name__.rpartition('.')
    prefix = before + sep
    registry = {}
    for _, modname, _ in pkgutil.walk_packages([path], prefix):
        submodule = importlib.import_module(modname, __name__)
        if hasattr(submodule, 'register'):
            submodule.register(registry)
        else:
            key = getattr(submodule, 'NAME', modname.split('.')[-1])
            if hasattr(submodule, 'notify_factory'):
                registry[key] = submodule.notify_factory
            elif hasattr(submodule, 'notify'):
                registry[key] = dummy_notify_factory(submodule.notify)
    return registry


class CompositeNotifier(object):

    REGISTRY = load_notifiers()

    def __init__(self, conf):
        self.conf = conf
        notifiers_conf = conf.get('notify', [])
        if not notifiers_conf:
            logger.warning(
                "No notifications configured for %r",
                conf['name'],
            )
        self.notifiers = []
        for notifier_conf in notifiers_conf:
            self.add_notifier(notifier_conf)

    def add_notifier(self, notifier_conf):
        try:
            name, value = next(iter(notifier_conf.items()))
        except AttributeError:
            name, value = notifier_conf, None
        try:
            notify_factory = self.REGISTRY[name]
        except KeyError:
            logger.error("Unknown notifier %r", name)
        else:
            self.notifiers.append(
                notify_factory(conf=self.conf, value=value)
            )

    def notify(self, report):
        if report:
            logger.debug('Sending report: %r', report)
            for notifier in self.notifiers:
                try:
                    notifier(report=report)
                except:
                    logger.exception(
                        "Exception occurred during sending notification"
                    )
        else:
            logger.debug('Report is empty, skipping notification')
    __call__ = notify



import contextlib
import tempfile
import shutil
from kibitzr.storage import PageHistory


@contextlib.contextmanager
def history(style=None):
    storage_dir = tempfile.mkdtemp()
    page_history = PageHistory(
        conf={'name': 'test', 'url': 'web'},
        storage_dir=storage_dir,
        style=style,
    )
    try:
        yield page_history
    finally:
        shutil.rmtree(storage_dir)


def test_unified_diff_sample():
    with history('default') as page_history:
        scenario = (
            (u"hello", True, u"test at web\n@@ -0,0 +1 @@\n+hello"),
            (u"world", True, u"test at web\n@@ -1 +1 @@\n-hello\n+world"),
            (u"world", False, None),
        )
        for content, changed, report in scenario:
            result = page_history.report_changes(content)
            assert result == (changed, report)


def test_verbose_diff_sample():
    with history('verbose') as page_history:
        scenario = (
            (u"hello", True, u"test at web\nhello"),
            (u"world", True, u"test at web\n"
                             u"New value:\nworld\n"
                             u"Old value:\nhello\n"),
            (u"world", False, None),
        )
        for content, changed, report in scenario:
            result = page_history.report_changes(content)
            assert result == (changed, report)


def test_word_diff_sample():
    with history('word') as page_history:
        scenario = (
            (u"hello", True, u"hello"),
            (u"world", True, u"[-hello-]{+world+}"),
            (u"world", False, None),
        )
        for content, changed, report in scenario:
            r_changed, r_content = page_history.report_changes(content)
            assert r_changed == changed
            if r_content:
                assert r_content.splitlines()[0] == report
            else:
                assert r_content == report



import pytest
from kibitzr.notifier import notify_factory

from ...compat import mock
from ...helpers import SettingsMock


@pytest.fixture()
def settings():
    """Override native settings singleton with empty one"""
    return SettingsMock.instance()


def test_mailgun_sample(settings):
    settings.creds.update({
        'mailgun': {
            'key': 'key',
            'domain': 'domain',
            'to': 'to',
        }
    })
    notify = notify_factory({
        'name': 'subject',
        'notify': [{'mailgun': {'to': 'me'}}]
    })
    with mock.patch.object(notify.notifiers[0], "session") as fake_session:
        notify('report')
        fake_session.post.assert_called_once_with(
            'https://api.mailgun.net/v3/domain/messages',
            data={
                'to': ['me'],
                'text': 'report',
                'from': 'Kibitzr <mailgun@domain>',
                'subject': 'Kibitzr update for subject',
            }
        )



from bs4 import BeautifulSoup

from kibitzr.transformer import transform_factory


def run_transform(key, value, content):
    pipeline = transform_factory({'transform': [{key: value}]})
    return pipeline(True, content)


def test_tag_selector():
    ok, content = run_transform('tag', 'a', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector():
    ok, content = run_transform('css', 'body h2.nav a#page-link', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector_all():
    ok, content = run_transform('css-all', 'div', HTML)
    assert ok is True
    assert prettify(content) == '\n'.join([
        '<div id="content">',
        ' Hello world!',
        '</div>',
        '',
        '<div class="footer">',
        ' Footer content',
        '</div>',
    ])


def test_xpath_selector():
    ok, content = run_transform('xpath', '//body/h2/a[@id="page-link"]', HTML)
    assert ok is True
    assert content.strip() == '<a href="page.html" id="page-link">Page</a>'


def test_extract_test():
    ok, content = run_transform('text', None, HTML)
    assert ok is True
    assert content.strip() == "\n".join([
        "Page",
        "Hello world!",
        "Footer content",
    ])


HTML = """
<html>
    <body>
        <h2 class="header nav">
            <a href="page.html" id="page-link">Page</a>
        </h2>
        <div id="content">
            Hello world!
        </div>
        <div class="footer">
            Footer content
        </div>
    </body>
</html>
"""


def prettify(html):
    return "\n".join(
        child.prettify()
        for child in BeautifulSoup(html, "lxml").html.body.children
    )



import logging
import traceback


logger = logging.getLogger(__name__)


def fetcher_factory(conf):
    if needs_firefox(conf):
        from .browser.fetcher import firefox_fetcher
        return URLFetcher(conf, firefox_fetcher)
    elif is_script(conf):
        from .script import fetch_by_script
        return ScriptFetcher(conf, fetch_by_script)
    else:
        from .simple import requests_fetcher
        return URLFetcher(conf, requests_fetcher(conf))


class URLFetcher(object):

    def __init__(self, conf, fetcher_func):
        self.conf = conf
        self.fetcher_func = fetcher_func

    def fetch(self):
        self.log_announcement()
        try:
            ok, content = self.fetcher_func(self.conf)
        except:
            logger.exception(
                "Exception occurred while fetching check"
            )
            ok, content = False, traceback.format_exc()
        return ok, content
    __call__ = fetch

    def log_announcement(self):
        logger.info(u"Fetching %s at %s",
                    self.conf['name'], self.conf['url'])


class ScriptFetcher(URLFetcher):

    def log_announcement(self):
        logger.info("Fetching %r using script",
                    self.conf['name'])


def is_script(conf):
    return all((
        'url' not in conf,
        'script' in conf,
    ))


def needs_firefox(conf):
    return any(
        conf.get(key)
        for key in ('delay', 'scenario', 'form')
    )



import pytest
from .target.server import start_server, stop_server
from kibitzr.fetcher import cleanup_fetchers
from ..compat import mock
from ..helpers import SettingsMock


server_address = None


@pytest.fixture(scope="session", autouse=True)
def target_website(request):
    global server_address
    server_process, server_address = start_server()
    request.addfinalizer(cleanup_fetchers)
    request.addfinalizer(lambda: stop_server(server_process))
    for module in ("browser.fetcher", "script"):
        patch_object = mock.patch(
            "kibitzr.fetcher.%s.settings" % module,
            return_value=SettingsMock()
        )
        patch_object.start()
        request.addfinalizer(patch_object.stop)


@pytest.fixture
def target():
    global server_address
    return server_address


@pytest.fixture
def not_found_conf(target):
    return {
        'name': '404',
        'url': "http://{0}:{1}/page-not-found".format(*target),
    }


@pytest.fixture
def json_conf(target):
    return {
        'name': 'Test page',
        'url': "http://{0}:{1}/api.json".format(*target),
        'transform': ['json'],
    }


@pytest.fixture
def html_text_conf(target):
    return {
        'name': 'Index',
        'url': "http://{0}:{1}/index.html".format(*target),
        'transform': ['text'],
    }


@pytest.fixture()
def python_script_conf():
    return {
        'name': 'Echo',
        'script': {'python': 'ok, content = True, "python"'},
    }



import contextlib

import pytest
from kibitzr.conf import (
    settings,
    ReloadableSettings,
    PlainYamlCreds,
    ConfigurationError,
)
from ..compat import mock


sample_conf = """
notifiers:
    slack:
        url: XXX

templates:
    teamcity:
        transforms:
            - text
            - changes: verbose
        scenario:
            login
        notify:
            - smtp: kibitzrrr@gmail.com
        period: 3600

scenarios:
    login: |
        driver.find_element_by_css_selector(".login").click()

checks:
    - name: Project Build Status
      url: http://teamcity/build/id
      template: teamcity
      period: 60

    - batch: WordPress {0} Plugin
      url-pattern: "http://wordpress/{0}"
      notify:
        - mailgun
      items:
        - A
        - B
"""

sample_creds = """
mailgun:
    url: ZZZ
"""


checks = [
    {
        'name': 'Project Build Status',
        'url': 'http://teamcity/build/id',
        'transforms': ['text', {'changes': 'verbose'}],
        'scenario': 'driver.find_element_by_css_selector(".login").click()\n',
        'notify': [{'smtp': 'kibitzrrr@gmail.com'}],
        'period': 60,
    },
    {
        'name': 'WordPress A Plugin',
        'url': "http://wordpress/A",
        'notify': ['mailgun'],
        'period': 300,
    },
    {
        'name': 'WordPress B Plugin',
        'url': "http://wordpress/B",
        'notify': ['mailgun'],
        'period': 300,
    },
]

creds = {
    'mailgun': {'url': 'ZZZ'},
}

notifiers = {
    'slack': {'url': 'XXX'},
}


@contextlib.contextmanager
def patch_source(klass, method, data):
    fake_file = mock.mock_open(read_data=data)
    with mock.patch.object(klass,
                           method,
                           fake_file,
                           create=True) as fake_method:
        yield fake_method


@contextlib.contextmanager
def patch_creds(data):
    with patch_source(PlainYamlCreds, 'open_creds', data) as fake_method:
        yield fake_method


@contextlib.contextmanager
def patch_conf(data):
    with patch_source(ReloadableSettings, 'open_conf', data) as fake_method:
        yield fake_method


def test_complex_conf_sample():
    with patch_conf(sample_conf):
        with patch_creds(sample_creds):
            conf = ReloadableSettings('::')
    assert conf.checks == checks
    assert conf.creds['mailgun']['url'] == 'ZZZ'


@mock.patch("kibitzr.conf.os.path.exists", return_value=False)
def test_missing_config_raises_configuration_error(exists):
    with pytest.raises(ConfigurationError):
        settings()


def test_reread():
    conf1 = (
        "checks:\n"
        "  - name: Name\n"
        "    url: URL\n"
    )
    conf2 = conf1 + "    period: 60\n"
    with patch_creds(""):
        with patch_conf(conf1):
            conf = ReloadableSettings('::')
            # Config didn't change:
            assert not conf.reread()
        with patch_conf(conf2):
            # Config changed:
            assert conf.reread()
    assert conf.checks == [{'name': 'Name', 'url': 'URL', 'period': 60}]


def test_name_from_url_population():
    conf = "checks: [{url: 'http://example.com/page_name'}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks[0]['name'] == 'http-example-com-page_name'


def test_unnamed_check():
    conf = "checks: [{period: 1}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks == [{
        'name': 'Unnamed check 1',
        'period': 1,
    }]


def test_period_parse():
    conf = "checks: [{period: 1 hour}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks[0]['period'] == 3600


def test_empty_period():
    conf = "checks: [{name: x}]"
    with patch_conf(conf):
        conf = ReloadableSettings('::')
    assert conf.checks[0]['period'] == 300



import json
import logging

from .utils import wrap_dummy, bake_parametrized


logger = logging.getLogger(__name__)


def pretty_json(text):
    json_dump = json.dumps(
        json.loads(text),
        indent=2,
        sort_keys=True,
        ensure_ascii=False,
        # encoding='utf-8',
    )
    return True, u'\n'.join([
        line.rstrip()
        for line in json_dump.splitlines()
    ])


def run_jq(query, text):
    from kibitzr.compat import sh
    jq = sh.Command("jq").bake('--monochrome-output', '--raw-output')
    logger.debug("Running jq query %s against %s", query, text)
    try:
        command = jq(query, _in=text)
        if not command.stderr:
            success, result = True, command.stdout.decode('utf-8')
        else:
            success, result = False, command.stderr.decode('utf-8')
    except sh.ErrorReturnCode as exc:
        logger.exception("jq failure")
        success, result = False, exc.stderr
    logger.debug("jq transform success: %r, content: %r",
                 success, result)
    return success, result


def register():
    return {
        'json': wrap_dummy(pretty_json),
        'jq': bake_parametrized(run_jq),
    }



import logging
import json

import six

from .utils import bake_parametrized
from .html import deep_recursion, SoupOps


logger = logging.getLogger(__name__)


def jinja_transform(code, content, conf):
    from jinja2 import Template, TemplateError
    html = LazyHTML(content)
    xml = LazyXML(content)
    context = {
        'conf': conf,
        'content': content,
        'lines': content.splitlines(),
        'json': LazyJSON(content),
        'css': html.css,
        'xpath': xml.xpath,
    }
    template = Template(code)
    template.environment.filters['text'] = text_filter
    try:
        return True, template.render(context)
    except TemplateError:
        logger.warning("Jinja transform failed", exc_info=True)
        return False, None


def text_filter(html):
    if isinstance(html, list):
        html = "".join(html)
    ok, content = SoupOps.extract_text(html)
    if ok:
        return content
    else:
        raise RuntimeError("Extract text failed")


class LazyJSON(object):
    def __init__(self, content):
        self.text = content
        self._json = None

    @property
    def json(self):
        if self._json is None:
            self._json = json.loads(self.text)
        return self._json

    def __getitem__(self, key):
        return self.json[key]


class LazyHTML(object):
    def __init__(self, content):
        self.html = content
        self._soup = None

    @property
    def soup(self):
        from bs4 import BeautifulSoup
        if self._soup is None:
            self._soup = BeautifulSoup(self.html, "html.parser")
        return self._soup

    def css(self, selector):
        with deep_recursion():
            elements = self.soup.select(selector)
            result = [six.text_type(x)
                      for x in elements]
            return result


class LazyXML(object):
    def __init__(self, content):
        from lxml import etree
        self.xml = content
        self._root = None
        self.etree = etree

    @property
    def root(self):
        if self._root is None:
            self._root = self.etree.fromstring(
                self.xml,
                parser=self.etree.HTMLParser(),
            )
        return self._root

    def xpath(self, selector):
        elements = self.root.xpath(selector)
        return [
            self.etree.tostring(
                element,
                method='html',
                pretty_print=True,
                encoding='unicode',
            )
            for element in elements
        ]


def register():
    return {
        'jinja': bake_parametrized(jinja_transform, pass_conf=True)
    }



import os
import logging
import tempfile


logger = logging.getLogger(__name__)


def execute_cmd(code, stdin=None):
    from kibitzr.compat import sh
    logger.info("Executing Batch script")
    logger.debug(code)
    if stdin is not None:
        if not stdin.strip():
            logger.info("Skipping execution with empty content")
            return True, stdin
        stdin = stdin
    with tempfile.NamedTemporaryFile() as fp:
        logger.debug("Saving code to %r", fp.name)
        fp.write(code.encode('utf-8'))
        fp.flush()
        logger.debug("Launching script %r", fp.name)
        try:
            result = sh.Command("cmd")("/Q", "/C", fp.name, _in=stdin)
            ok = True
        except sh.ErrorReturnCode as exc:
            result = exc
            ok = False
    stdout = result.stdout
    stderr = result.stderr
    if ok:
        log = logger.debug
        report = stdout
    else:
        log = logger.error
        report = stderr
    log("Bash exit_code: %r", result.exit_code)
    log("Bash stdout: %s", stdout)
    log("Bash stderr: %s", stderr)
    return ok, report



import logging
import traceback

from ..conf import settings
from ..bash import execute_bash


logger = logging.getLogger(__name__)
PYTHON_ERROR = "script.python must set global variables ok and content"


def fetch_by_script(conf):
    code = conf['script']
    try:
        python_code = code['python']
    except (KeyError, TypeError):
        # Not a python script
        pass
    else:
        return fetch_by_python(python_code, conf)
    try:
        # Explicit notation:
        bash_script = code['bash']
    except (KeyError, TypeError):
        bash_script = code
    return execute_bash(bash_script)


def fetch_by_python(code, conf):
    logger.info("Fetch using Python script")
    logger.debug(code)
    assert 'content' in code, PYTHON_ERROR
    try:
        # ok, content = False, None
        namespace = {'ok': True}
        exec(code, {'conf': conf, 'creds': settings().creds}, namespace)
        return namespace['ok'], namespace['content']
    except:
        logger.exception("Python fetcher raised an Exception")
        return False, traceback.format_exc()



import logging

from .fetcher import fetcher_factory
from .notifier import notify_factory
from .transformer import transform_factory


logger = logging.getLogger(__name__)


class Checker(object):
    def __init__(self, conf):
        self.conf = conf
        self.fetch = fetcher_factory(conf)
        self.transform = transform_factory(self.conf)
        self.notify = notify_factory(self.conf)

    @classmethod
    def create_from_settings(cls, checks, names=None):
        if names:
            selected_checks = [
                conf
                for conf in checks
                if conf['name'] in names
            ]
            selected_names = [conf['name'] for conf in selected_checks]
            if len(selected_checks) < len(checks):
                logger.info("Filtered list of checks to: %r",
                            ", ".join(sorted(selected_names)))
                checks = selected_checks
            if len(selected_names) < len(names):
                logger.error(
                    "Following check(s) were not found: %r",
                    ", ".join(sorted(set(names).difference(selected_names)))
                )
        return [cls(conf)
                for conf in checks]

    def check(self):
        ok, content = self.fetch()
        ok, report = self.transform(ok, content)
        self.notify(report=report)
        return ok, report



from ..bash import execute_bash


class BashNotify(object):

    def __init__(self, value):
        self.code = value

    def __call__(self, report):
        execute_bash(self.code, report)


def notify_factory(conf, value):
    return BashNotify(value)



from kibitzr.transformer.jinja_transform import jinja_transform


def test_content_is_passed():
    ok, content = jinja_transform('hello {{ content }}!', 'world', {})
    assert ok is True
    assert content == "hello world!"


def test_lines_are_passed():
    ok, content = jinja_transform('{{ lines[1] }}', 'a\nb', {})
    assert ok is True
    assert content == "b"


def test_json_is_passed():
    ok, content = jinja_transform('{{ json["a"][1] }}', '{"a": [1, 2, 3]}', {})
    assert ok is True
    assert content == "2"


def test_css_selector_is_passed():
    ok, content = jinja_transform(
        '{{ css("div p") | text }}',
        '<div><a>A</a><p>P</p></div>',
        {},
    )
    assert ok is True
    assert content == "P"


def test_xpath_selector_is_passed():
    ok, content = jinja_transform(
        '{{ xpath("//div/p") | text }}',
        '<div><a>A</a><p>P</p></div>',
        {},
    )
    assert ok is True
    assert content == "P"



import functools


def wrap_dummy(transform_func):
    def transform_factory(value, conf):
        return transform_func
    return transform_factory


def bake_parametrized(transform_func, pass_conf=False, **kwargs):
    def transform_factory(value, conf):
        if pass_conf:
            return functools.partial(
                transform_func,
                value,
                conf=conf,
                **kwargs
            )
        else:
            return functools.partial(
                transform_func,
                value,
                **kwargs
            )
    return transform_factory



from .webhook import WebHookNotify, webhook_factory


class ZapierNotify(WebHookNotify):
    CREDS_KEY = "zapier"
    POST_KEY = "text"

    def configure_session(self):
        self.session.headers.update({
            "Content-Type": "application/x-www-form-urlencoded"
        })


notify_factory = webhook_factory(ZapierNotify)



from kibitzr.transformer.plain_text import (
    python_transform,
    bash_transform,
)


def test_bash_transform_sample():
    ok, content = bash_transform(
        code="sed 's/A/B/g'",
        content="ACTGA",
    )
    assert ok is True
    assert content == "BCTGB"


def test_python_transform_sample():
    ok, content = python_transform(
        code="content = content.replace('A', 'B')",
        content="ACTGA",
        conf=None,
    )
    assert ok is True
    assert content == "BCTGB"


def test_bash_transform_error_is_captured():
    ok, content = bash_transform(
        code="ls /NO-SUCH-DIR",
        content="?",
    )
    assert ok is False
    assert "ls: cannot access" in content
    assert "/NO-SUCH-DIR" in content
    assert "No such file or directory" in content


def test_python_exception_is_captured():
    ok, content = python_transform(
        code="content = 1 / 0",
        content="?",
        conf=None,
    )
    assert ok is False
    assert content.splitlines()[-1].startswith("ZeroDivisionError")



import os
import itertools
import io

from kibitzr.compat import sh

from .utils import normalize_filename


def report_changes(conf, content):
    return PageHistory(conf).report_changes(content)


class PageHistory(object):
    """
    Single file changes history using git
    """
    STORAGE_DIR = "pages"

    def __init__(self, conf, storage_dir=None, style=None):
        self.storage_dir = storage_dir or self.STORAGE_DIR
        self.cwd = os.path.join(
            self.storage_dir,
            normalize_filename(conf['name']),
        )
        self.target = os.path.join(self.cwd, "content")
        self.git = sh.Command('git').bake(
            '--no-pager',
            _cwd=self.cwd,
        )
        self.ensure_repo_exists()
        if conf.get('url'):
            self.commit_msg = u"{name} at {url}".format(
                name=conf['name'],
                url=conf.get('url'),
            )
        else:
            self.commit_msg = conf['name']
        self.reporter = ChangesReporter(
            self.git,
            self.commit_msg,
            style,
        )

    def report_changes(self, content):
        """
        1) Write changes in file,
        2) Commit changes in git
        3.1) If something changed, return tuple(True, changes)
        3.2) If nothing changed, return tuple(False, None)
        If style is "verbose", return changes in human-friendly format,
        else use unified diff
        """
        self.write(content)
        if self.commit():
            return True, self.reporter.report()
        else:
            return False, None

    def write(self, content):
        """Save content on disk"""
        with io.open(self.target, 'w', encoding='utf-8') as fp:
            fp.write(content)
            if not content.endswith(u'\n'):
                fp.write(u'\n')

    def commit(self):
        """git commit and return whether there were changes"""
        self.git.add('-A', '.')
        try:
            self.git.commit('-m', self.commit_msg)
            return True
        except sh.ErrorReturnCode_1:
            return False

    def ensure_repo_exists(self):
        """Create git repo if one does not exist yet"""
        if not os.path.isdir(self.cwd):
            os.makedirs(self.cwd)
        if not os.path.isdir(os.path.join(self.cwd, ".git")):
            self.git.init()
            self.git.config("user.email", "you@example.com")
            self.git.config("user.name", "Your Name")


class ChangesReporter(object):

    def __init__(self, git, subject, style=None):
        self.git = git
        self.subject = subject
        self.report = getattr(self, style or 'default', self.default)

    def word(self):
        """Return last changes with word diff"""
        try:
            output = self.git.diff(
                '--no-color',
                '--word-diff=plain',
                'HEAD~1:content',
                'HEAD:content',
            ).stdout.decode('utf-8')
        except sh.ErrorReturnCode_128:
            result = self.git.show(
                "HEAD:content").stdout.decode("utf-8")
        else:
            ago = self.git.log(
                '-2',
                '--pretty=format:last change was %cr',
                'content'
            ).stdout.decode('utf-8').splitlines()
            lines = output.splitlines()
            result = u'\n'.join(
                itertools.chain(
                    itertools.islice(
                        itertools.dropwhile(
                            lambda x: not x.startswith('@@'),
                            lines[1:],
                        ),
                        1,
                        None,
                    ),
                    itertools.islice(ago, 1, None),
                )
            )
        return result

    def default(self):
        """Return last changes in truncated unified diff format"""
        output = self.git.log(
            '-1',
            '-p',
            '--no-color',
            '--format=%s',
        ).stdout.decode('utf-8')
        lines = output.splitlines()
        return u'\n'.join(
            itertools.chain(
                lines[:1],
                itertools.islice(
                    itertools.dropwhile(
                        lambda x: not x.startswith('+++'),
                        lines[1:],
                    ),
                    1,
                    None,
                ),
            )
        )

    def verbose(self):
        """Return changes in human-friendly format #14"""
        try:
            before = self.git.show('HEAD~1:content').strip()
        except sh.ErrorReturnCode_128:
            before = None
        after = self.git.show('HEAD:content').strip()
        if before is not None:
            return (u'{subject}\nNew value:\n{after}\n'
                    u'Old value:\n{before}\n'
                    .format(subject=self.subject,
                            before=before,
                            after=after))
        else:
            return u'\n'.join([self.subject, after])



from __future__ import absolute_import
import logging

from ..conf import settings


logger = logging.getLogger(__name__)


class TelegramBot(object):
    def __init__(self, chat_id=None):
        from telegram.bot import Bot
        telegram_creds = settings().creds['telegram']
        token = telegram_creds['token']
        if chat_id is not None:
            self._chat_id = chat_id
        else:
            self._chat_id = telegram_creds.get('chat')
        self.bot = Bot(token=token)

    @property
    def chat_id(self):
        if self._chat_id is None:
            chat = self.bot.getUpdates(limit=1)[0].message.chat
            logger.debug("Imprinted chat id %d of type %s",
                         chat.id, chat.type)
            self._chat_id = chat.id
        return self._chat_id

    def post(self, report, **kwargs):
        message = self.bot.send_message(self.chat_id, report)
        return message

    __call__ = post


def notify_factory(conf, value):
    return TelegramBot(value).post


def chat_id():
    bot = TelegramBot()
    print(bot.chat_id)



import sys
import contextlib
import logging
import functools

import six

from .utils import bake_parametrized


logger = logging.getLogger(__name__)


class SoupOps(object):
    def __init__(self, selector=None, select_all=False):
        self.selector = selector
        self.select_all = select_all

    def tag_selector(self, html):
        with soup(html) as doc:
            element = doc.find(self.selector)
            if element:
                return True, six.text_type(element)
            else:
                logger.warning('Tag not found: %r', self.selector)
                return False, html

    def css_selector(self, html):
        with soup(html) as doc:
            try:
                elements = doc.select(self.selector)
                if self.select_all:
                    result = u"".join(six.text_type(x)
                                      for x in elements)
                else:
                    result = six.text_type(elements[0])
                return True, result
            except IndexError:
                logger.warning('CSS selector not found: %r', self.selector)
                return False, html

    @staticmethod
    def extract_text(html):
        with soup(html) as doc:
            strings = doc.stripped_strings
            return True, u'\n'.join([
                line
                for line in strings
                if line
            ])

    @classmethod
    def factory(cls, key, value, conf):
        def transform(content):
            instance = cls(selector=value, select_all=select_all)
            method = handler.__get__(instance, cls)
            return method(content)
        action, _, all_flag = key.partition('-')
        select_all = (all_flag == 'all')
        handler = cls.SHORTCUTS[action]
        return transform

    SHORTCUTS = {
        'tag': tag_selector,
        'css': css_selector,
        'text': extract_text,
    }


def xpath_selector(selector, html):
    from lxml import etree
    root = etree.fromstring(html, parser=etree.HTMLParser())
    elements = root.xpath(selector)
    if elements:
        return True, etree.tostring(
            next(iter(elements)),
            method='html',
            pretty_print=True,
            encoding='unicode',
        )
    else:
        logger.warning('XPath selector not found: %r', selector)
        return False, html


@contextlib.contextmanager
def soup(html):
    from bs4 import BeautifulSoup
    with deep_recursion():
        yield BeautifulSoup(html, "html.parser")


@contextlib.contextmanager
def deep_recursion():
    old_limit = sys.getrecursionlimit()
    try:
        sys.setrecursionlimit(100000)
        yield
    finally:
        sys.setrecursionlimit(old_limit)


def bake_html(key):
    return functools.partial(SoupOps.factory, key)


def register():
    """
    Return dictionary of tranform factories
    """
    registry = {
        key: bake_html(key)
        for key in ('css', 'css-all', 'tag', 'text')
    }
    registry['xpath'] = bake_parametrized(xpath_selector)
    return registry



import pytest

from kibitzr.fetcher.script import fetch_by_script


def test_fetch_by_script_default_is_bash():
    ok, content = fetch_by_script(
        {'script': 'echo "hello"'}
    )
    assert ok is True
    assert content == "hello\n"


def test_fetch_by_explicit_bash_script():
    ok, content = fetch_by_script(
        {'script': {'bash': 'echo "hello"'}}
    )
    assert ok is True
    assert content == "hello\n"


def test_fetch_by_python_script():
    ok, content = fetch_by_script(
        {'script': {'python': 'ok, content = True, "hello"'}}
    )
    assert ok is True
    assert content == "hello"


def test_python_script_must_define_content():
    with pytest.raises(AssertionError):
        fetch_by_script(
            {'script': {'python': 'dummy'}}
        )


def test_python_script_exception_is_reported():
    ok, content = fetch_by_script(
        {'script': {'python': 'content = 1 / 0'}}
    )
    assert ok is False
    assert content.splitlines()[-1].startswith("ZeroDivisionError")


def test_ok_is_optional_in_python_script():
    ok, content = fetch_by_script(
        {'script': {'python': 'content = "dummy"'}}
    )
    assert ok is True
    assert content == "dummy"


def test_bash_error_is_returned():
    ok, content = fetch_by_script(
        {'script': 'no command not found'}
    )
    assert ok is False
    assert content.split(':', 1)[1].strip() == "line 1: no: command not found"



import sys
import logging

import click


LOG_LEVEL_CODES = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


@click.group()
@click.option("-l", "--log-level", default="info",
              type=click.Choice(LOG_LEVEL_CODES.keys()),
              help="Logging level")
@click.pass_context
def cli(ctx, log_level):
    """Run kibitzr COMMAND --help for detailed descriptions"""
    ctx.obj = {'log_level': LOG_LEVEL_CODES[log_level.lower()]}


@cli.command()
def version():
    """Print version"""
    from kibitzr import __version__ as kibitzr_version
    print(kibitzr_version)


@cli.command()
def firefox():
    """Launch Firefox with persistent profile"""
    from kibitzr.main import run_firefox
    run_firefox()


@cli.command()
@click.argument('name', nargs=-1)
@click.pass_context
def once(ctx, name):
    """Run kibitzr checks once and exit"""
    from kibitzr.main import main
    sys.exit(main(once=True, log_level=ctx.obj['log_level'], names=name))


@cli.command()
@click.argument('name', nargs=-1)
@click.pass_context
def run(ctx, name):
    """Run kibitzr in the foreground mode"""
    from kibitzr.main import main
    sys.exit(main(once=False, log_level=ctx.obj['log_level'], names=name))


@cli.command()
def init():
    """Create boilerplate configuration files"""
    from kibitzr.main import bootstrap
    bootstrap()


@cli.command()
def telegram_chat():
    """Return chat id for the last message sent to Telegram Bot"""
    from kibitzr.main import telegram_chat
    telegram_chat()


if __name__ == "__main__":
    cli()



import os
import stat
import logging


logger = logging.getLogger(__name__)


KIBITZR_YML = """
# Format: http://kibitzr.readthedocs.io/en/latest/configuration.html
checks:
  - name: Kibitzr release
    url: https://pypi.python.org/pypi/kibitzr/json
    transform:
      - jinja: |
          {{ json["info"]["version"] }}
      - changes: verbose
    notify:
      - python: print(content)
    period: 5
""".lstrip()


KIBITZR_CREDS_YML = """
# Plain text credentials configuration
# Documentation and other options: http://kibitzr.readthedocs.io/en/latest/credentials.html
service:
  username: john
  password: doe
""".lstrip()


def create_boilerplate():
    """
    Create kibitzr.yml and kibitzr-creds.yml in current directory
    if they do not exist.
    """
    if not os.path.exists('kibitzr.yml'):
        with open('kibitzr.yml', 'wt') as fp:
            logger.info("Saving sample check in kibitzr.yml")
            fp.write(KIBITZR_YML)
    else:
        logger.info("kibitzr.yml already exists. Skipping")
    if not os.path.exists('kibitzr-creds.yml'):
        with open('kibitzr-creds.yml', 'wt') as fp:
            logger.info("Creating kibitzr-creds.yml")
            fp.write(KIBITZR_CREDS_YML)
        os.chmod('kibitzr-creds.yml', stat.S_IRUSR | stat.S_IWUSR)
    else:
        logger.info("kibitzr-creds.yml already exists. Skipping")



from kibitzr.checker import Checker


def test_fill_form_sample(target):
    conf = {
        'name': 'Test page',
        'url': "http://{0}:{1}/form.html".format(*target),
        'form': [
            {'id': 'name', 'value': '{{ "name" | sort | join("") }}'},
            {'css': '#pass', 'creds': 'pass'}
        ],
        # 'scenario': 'import pdb; pdb.set_trace()',
        'transform': [{'css': '#params'}, 'text'],
        # 'headless': False,
    }
    ok, content = Checker(conf).check()
    assert ok is True
    assert content == "\n".join([
        "name = aemn",
        "pass = password",
    ])



"""
FireFox housekeeping - starting and stopping process
"""

import os
import sys
import shutil
import logging
from contextlib import contextmanager


PROFILE_DIR = 'firefox_profile'
logger = logging.getLogger(__name__)
FIREFOX_INSTANCE = {
    'xvfb_display': None,
    'driver': None,
    'headed_driver': None,
}


def cleanup():
    """Must be called before exit"""
    temp_dirs = []
    if FIREFOX_INSTANCE['driver'] is not None:
        if FIREFOX_INSTANCE['driver'].profile:
            temp_dirs.append(FIREFOX_INSTANCE['driver'].profile.profile_dir)
        FIREFOX_INSTANCE['driver'].quit()
        FIREFOX_INSTANCE['driver'] = None
    if FIREFOX_INSTANCE['headed_driver'] is not None:
        if FIREFOX_INSTANCE['headed_driver'].profile:
            temp_dirs.append(FIREFOX_INSTANCE['headed_driver'].profile.profile_dir)
        FIREFOX_INSTANCE['headed_driver'].quit()
        FIREFOX_INSTANCE['headed_driver'] = None
    if FIREFOX_INSTANCE['xvfb_display'] is not None:
        FIREFOX_INSTANCE['xvfb_display'].stop()
        FIREFOX_INSTANCE['xvfb_display'] = None
    for temp_dir in temp_dirs:
        shutil.rmtree(temp_dir, ignore_errors=True)


@contextmanager
def firefox(headless=True):
    """
    Context manager returning Selenium webdriver.
    Instance is reused and must be cleaned up on exit.
    """
    from selenium import webdriver
    from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
    if headless:
        if FIREFOX_INSTANCE['xvfb_display'] is None:
            FIREFOX_INSTANCE['xvfb_display'] = virtual_buffer()
        driver_key = 'driver'
    else:
        driver_key = 'headed_driver'
    if FIREFOX_INSTANCE[driver_key] is None:
        if logger.level == logging.DEBUG:
            firefox_binary = FirefoxBinary(log_file=sys.stdout)
        else:
            firefox_binary = None
        # Load profile, if it exists:
        if os.path.isdir(PROFILE_DIR):
            firefox_profile = webdriver.FirefoxProfile(PROFILE_DIR)
        else:
            firefox_profile = None
        FIREFOX_INSTANCE[driver_key] = webdriver.Firefox(
            firefox_binary=firefox_binary,
            firefox_profile=firefox_profile,
        )
        FIREFOX_INSTANCE[driver_key].set_window_size(1024, 768)
    yield FIREFOX_INSTANCE[driver_key]


def virtual_buffer():
    """
    Try to start xvfb, trying multiple (up to 5) times if a failure
    """
    from xvfbwrapper import Xvfb
    for _ in range(0, 6):
        xvfb_display = Xvfb()
        xvfb_display.start()
        # If Xvfb started, return.
        if xvfb_display.proc is not None:
            return xvfb_display
    raise Exception("Xvfb could not be started after six attempts.")



import os
import re
import copy
import logging.config
import contextlib

import six
import yaml
import pytimeparse
import entrypoints


logger = logging.getLogger(__name__)


class ConfigurationError(RuntimeError):
    pass


class ReloadableSettings(object):
    _instance = None
    CONFIG_DIRS = (
        '',
        '~/.config/kibitzr/',
        '~/',
    )
    CONFIG_FILENAME = 'kibitzr.yml'
    RE_PUNCTUATION = re.compile(r'\W+')
    UNNAMED_PATTERN = 'Unnamed check {0}'

    def __init__(self, config_dir):
        self.filename = os.path.join(config_dir, self.CONFIG_FILENAME)
        self.checks = None
        self.creds = CompositeCreds(config_dir)
        self.parser = SettingsParser()
        self.reread()

    @classmethod
    def detect_config_dir(cls):
        candidates = [
            (directory, os.path.join(directory, cls.CONFIG_FILENAME))
            for directory in map(os.path.expanduser, cls.CONFIG_DIRS)
        ]
        for directory, file_path in candidates:
            if os.path.exists(file_path):
                return directory
        raise ConfigurationError(
            "kibitzr.yml not found in following locations: %s"
            % ", ".join([x[1] for x in candidates])
        )

    @classmethod
    def instance(cls):
        if cls._instance is None:
            config_dir = cls.detect_config_dir()
            cls._instance = cls(config_dir)
        return cls._instance

    def reread(self):
        """
        Read configuration file and substitute references into checks conf
        """
        logger.debug("Loading settings from %s",
                     os.path.abspath(self.filename))
        conf = self.read_conf()
        changed = self.creds.reread()
        checks = self.parser.parse_checks(conf)
        if self.checks != checks:
            self.checks = checks
            return True
        else:
            return changed

    def read_conf(self):
        """
        Read and parse configuration file
        """
        with self.open_conf() as fp:
            return yaml.load(fp)

    @contextlib.contextmanager
    def open_conf(self):
        with open(self.filename) as fp:
            yield fp


class CompositeCreds(object):

    def __init__(self, config_dir):
        self.plain = PlainYamlCreds(config_dir)
        self.extensions = {}
        self.load_extensions()

    def reread(self):
        changed = False
        for extension in self.extensions:
            reread_method = getattr(extension, 'reread', None)
            if reread_method:
                changed |= reread_method()
        return changed

    def load_extensions(self):
        for point in entrypoints.get_group_all("kibitzr.creds"):
            factory = point.load()
            self.extensions[point.name] = factory()

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        if key in self.extensions:
            return self.extensions[key]
        elif key in self.plain:
            return self.plain[key]
        else:
            raise KeyError("Credentials not found: {0}".format(key))


class PlainYamlCreds(object):

    CREDENTIALS_FILENAME = 'kibitzr-creds.yml'

    def __init__(self, config_dir):
        super(PlainYamlCreds, self).__init__()
        self.creds = {}
        self.creds_filename = os.path.join(config_dir,
                                           self.CREDENTIALS_FILENAME)
        self.reread()

    def __contains__(self, key):
        return key in self.creds

    def __getitem__(self, key):
        return self.creds[key]

    def reread(self):
        """
        Read and parse credentials file.
        If something goes wrong, log exception and continue.
        """
        logger.debug("Loading credentials from %s",
                     os.path.abspath(self.creds_filename))
        creds = {}
        try:
            with self.open_creds() as fp:
                creds = yaml.load(fp)
        except IOError:
            logger.info("No credentials file found at %s",
                        os.path.abspath(self.creds_filename))
        except:
            logger.exception("Error loading credentials file")
        if creds != self.creds:
            self.creds = creds
            return True
        return False

    @contextlib.contextmanager
    def open_creds(self):
        with open(self.creds_filename) as fp:
            yield fp


def settings():
    """
    Returns singleton instance of settings
    """
    return ReloadableSettings.instance()


class SettingsParser(object):
    RE_PUNCTUATION = re.compile(r'\W+')
    UNNAMED_PATTERN = 'Unnamed check {0}'

    def parse_checks(self, conf):
        """
        Unpack configuration from human-friendly form
        to strict check definitions.
        """
        checks = conf.get('checks', conf.get('pages', []))
        checks = list(self.unpack_batches(checks))
        checks = list(self.unpack_templates(checks, conf.get('templates', {})))
        self.inject_missing_names(checks)
        for check in checks:
            self.inject_scenarios(check, conf.get('scenarios', {}))
            self.inject_notifiers(check, conf.get('notifiers', {}))
            self.fix_period(check)
        return checks

    @staticmethod
    def inject_notifiers(check, notifiers):
        if 'notify' in check:
            for notify in check['notify']:
                if hasattr(notify, 'keys'):
                    notify_type = next(iter(notify.keys()))
                    notify_param = next(iter(notify.values()))
                    try:
                        notify[notify_type] = notifiers[notify_param]
                    except (TypeError, KeyError):
                        # notify_param is not a predefined notifier name
                        # Save it as is:
                        notify[notify_type] = notify_param

    @staticmethod
    def inject_scenarios(check, scenarios):
        try:
            shared_scenario = scenarios[check['scenario']]
        except (KeyError, TypeError):
            pass
        else:
            check['scenario'] = shared_scenario

    @staticmethod
    def fix_period(check):
        period = check.setdefault('period', 300)
        if isinstance(period, six.string_types):
            seconds = int(pytimeparse.parse(period))
            logger.debug('Parsed "%s" to %d seconds',
                         period, seconds)
            check['period'] = seconds

    @staticmethod
    def unpack_batches(checks):
        for check in checks:
            if 'batch' in check:
                base = copy.deepcopy(check)
                batch = base.pop('batch')
                url_pattern = base.pop('url-pattern')
                items = base.pop('items')
                for item in items:
                    yield dict(
                        copy.deepcopy(base),
                        name=batch.format(item),
                        url=url_pattern.format(item),
                    )
            else:
                yield check

    @classmethod
    def inject_missing_names(cls, checks):
        unnamed_check_counter = 1
        for check in checks:
            if not check.get('name'):
                if check.get('url'):
                    check['name'] = cls.url_to_name(check['url'])
                else:
                    check['name'] = cls.UNNAMED_PATTERN.format(unnamed_check_counter)
                    unnamed_check_counter += 1

    @classmethod
    def url_to_name(cls, url):
        return cls.RE_PUNCTUATION.sub('-', url)

    @staticmethod
    def unpack_templates(checks, templates):
        for check in checks:
            if 'template' in check:
                if check['template'] in templates:
                    templated_check = dict(
                        copy.deepcopy(templates[check['template']]),
                        **check
                    )
                    del templated_check['template']
                    yield templated_check
                else:
                    raise ConfigurationError(
                        "Template %r not found. Referenced in check %r"
                        % (check['template'], check['name'])
                    )
            else:
                yield check


logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True,
        },
        'sh': {
            'level': 'INFO',
        },
        'sh.command': {
            'level': 'WARNING',
        },
    }
})



from kibitzr.conf import ReloadableSettings


class SettingsMock(ReloadableSettings):

    def __init__(self):
        self.checks = []
        self.notifiers = {}
        self.creds = {'pass': 'password'}

    @classmethod
    def instance(cls):
        ReloadableSettings._instance = cls()
        return ReloadableSettings._instance
